#ifndef _profile_cholesky_c /* Copyright Bill.Triggs@inrialpes.fr */
#define _profile_cholesky_c "$ Sun Jan 18 18:53:45 2004 $"
/** NO WARRANTY. Use & distribute freely. Fixes/feedback welcome. **/
/*------------------------------------------------------------------------
  Profile (skyline) Cholesky decomposition of a positive definite matrix A.
  To solve A*x=b, use the following steps:

  profile_cholesky(L,A,n,cmin,lambda);	-- return L with L*L'=A
  x=b;
  profile_chol_fwsub(L,n,cmin,x);	-- return x with L*x=b
  profile_chol_bksub(L,n,cmin,x);	-- return x with L'*x=b  
 ------------------------------------------------------------------------*/
#ifndef _math_h
# include <math.h>
#endif
#ifndef _profile_matrix_h
# include "profile_matrix.h"
#endif

#ifndef MAX
# define MAX(a,b)	((a)<(b)?(b):(a))
#endif
#ifndef FLOAT
# define FLOAT double
#endif
#ifndef SQ
# define SQ(x) ((x)*(x))
#endif

/*--------------------------------------------------------------------
  Forwards Choleski decomposition of a positive semi-definite profile
  matrix A, writing the result into the profile matrix L: A = L*L'.
  Only the lower triangle of A is accessed. L is lower triangular with
  the same profile as A, except that within-band zeros in A are
  usually filled in in L. A and L can point to the same matrix - if
  so, the algorithm operates in place. The decomposition is stabilized
  by rounding any small or negative diagonal elements up to lambda.
  The complexity is O(n^3/6) additions and multiplications if A is
  dense, and O(n.(band width)^2/2) if A is a narrow band matrix.
 --------------------------------------------------------------------*/
void
profile_cholesky(FLOAT* l[/*n*/], FLOAT*const a[/*n*/], int n, 
		 const int cmin[/*n*/], FLOAT lambda)
{
  double t, lambda2 = SQ(lambda);
  int i,j,k;
  for (i=0; i<n; i++) {
     const FLOAT *ai=a[i];
     FLOAT *li=l[i];
     int cmini=cmin[i];
     for (j=cmini; j<i; j++) {
	FLOAT* lj=l[j];
	for (k=MAX(cmini,cmin[j]), t=ai[j]; k<j; k++) 
	   t -= li[k]*lj[k];
	li[j] = t/lj[j];
     }
     for (k=cmini, t=ai[i]; k<i; k++) 
	t -= SQ(li[k]);
     li[i] = (t>lambda2)? sqrt(t) : lambda;
  }
}

/*------------------------------------------------------------------------
  Forwards substitution to solve L*x = b for x, for a lower triangular
  profile matrix L. The algorithm receives b in x and operates in place.
 ------------------------------------------------------------------------*/
void
profile_chol_fwsub(FLOAT*const l[/*n*/], int n, const int cmin[/*n*/], 
		   FLOAT x[/*n*/])
{
  int i,j;
  for (i=0; i<n; i++) {
     const FLOAT *li=l[i];
     double t=x[i];
     for (j=cmin[i]; j<i; j++)
	t -= li[j]*x[j];
     x[i] = t/li[i];
  }
}

/*--------------------------------------------------------------------
  Backwards substitution to solve the transposed linear system L'*x=b
  for x, for a lower triangular profile matrix L.  The algorithm
  receives b in x and operates in place.
 --------------------------------------------------------------------*/
void
profile_chol_bksub(FLOAT*const l[/*n*/], int n, const int cmin[/*n*/], 
		   FLOAT x[/*n*/])
{
   int i,j;
   for (i=n-1; i>=0; i--) {
      const FLOAT *li = l[i];
      double xi = x[i]/li[i];
      for (j=cmin[i]; j<i; j++)
	 x[j] -= li[j]*xi;
      x[i] = xi;
   }
}

#ifdef TEST
/*--------------------------------------------------------------------
  Test routine for profile Cholesky methods. Chooses a random profile,
  builds a random symmetric positive definite test matrix with this
  profile, decomposes it and checks the decomposition, solves and
  checks random forwards, backwards and forwards-backwards
  substitutions.  Uses integer random numbers to get exact solutions.
 --------------------------------------------------------------------*/
#include <stdio.h>
#include "profile_matrix.c"
#define N 100
#define N_ITER 10
#define DRAND() (double)(rand()%1000-500)
#define RAND() rand()

int
main()
{
   int n=N, nit=N_ITER, it,i,j,k, cmin[N];
   FLOAT **a, **l, **u, x[N], b[N];
   double s;

   /* Try a few test matrices.
    */
   for (it=0; it<nit; it++) {

      /* Build a random profile cmin.
       */
      for (i=0; i<n; i++) {
	 cmin[i] = RAND()%(i+1);
      }
      /* Allocate matrices with profile cmin.
       */
      a = new_profile_matrix(FLOAT,n,cmin,0);
      l = new_profile_matrix(FLOAT,n,cmin,0);
      u = new_profile_matrix(FLOAT,n,cmin,0);

      /* Build a random lower triangular matrix with profile cmin.
       */
      for (i=0; i<n; i++) {
	 for (j=cmin[i]; j<i; j++)
	    u[i][j] = DRAND();
	 u[i][i] = fabs(DRAND());
      }
      /* Form A = U*U'.
       */
      for (i=0; i<n; i++) {
	 for (j=cmin[i]; j<=i; j++) {
	    double t = 0.0;
	    for (k=MAX(cmin[i],cmin[j]); k<=j; k++)
	       t += u[i][k]*u[j][k];
	    a[i][j] = t;
	 }
      }
      /* Cholesky decompose A, and check that we get U back.
       */
      profile_cholesky(l,a,n,cmin,0.0);

      for (i=0, s=0.0; i<n; i++) {
	 for (j=cmin[i]; j<=i; j++) 
	    s += fabs(l[i][j]-u[i][j]);
      }
      fprintf(stdout,"L*L'=A, residual %g\n",s);

      /* Build a random vector x.
       */
      for (i=0; i<n; i++) {
	 x[i] = DRAND();
      }

      /* Form b=L*x.
       */
      for (i=0; i<n; i++) {
	 double t = 0.0;
	 for (j=cmin[i]; j<=i; j++) {
	    t += l[i][j]*x[j];
	 }
	 b[i] = t;
      }
      /* Solve L*x=b and check that we get x back.
       */
      profile_chol_fwsub(l,n,cmin,b);
      for (i=0, s=0.0; i<n; i++) {
	 s += fabs(b[i]-x[i]);
      }
      fprintf(stdout,"L*x=b, residual %g\n",s);   

      /* Form b=L'*x.
       */
      for (i=0; i<n; i++) {
	 b[i] = 0.0;
	 for (j=cmin[i]; j<=i; j++) {
	    b[j] += l[i][j]*x[i];
	 }
      }
      /* Solve L*x=b and check that we get x back.
       */
      profile_chol_bksub(l,n,cmin,b);
      for (i=0, s=0.0; i<n; i++) {
	 s += fabs(b[i]-x[i]);
      }
      fprintf(stdout,"L'*x=b, residual %g\n",s);   

      /* Form b=A*x.
       */
      for (i=0; i<n; i++) {
	 double t = a[i][i]*x[i];
	 for (j=cmin[i]; j<i; j++) {
	    t += a[i][j]*x[j]; /* contribution from lower triangle */
	    b[j] += a[i][j]*x[i]; /* contribution from upper triangle */
	 }
	 b[i] = t;
      }

      /* Solve A*x=b, and check that we get x back.
       */
      profile_chol_fwsub(l,n,cmin,b);
      profile_chol_bksub(l,n,cmin,b);
      for (i=0, s=0.0; i<n; i++) {
	 s += fabs(b[i]-x[i]);
      }
      fprintf(stdout,"A*x=b, residual %g\n",s);   
   
      /* Free matrices.
       */
      free_profile_matrix(a);
      free_profile_matrix(u);
      free_profile_matrix(l);
   }
   
   return 0;
}

#endif /* TEST */
#endif /*_profile_cholesky_c */
